from __future__ import annotations

import json
from pathlib import PosixPath
from typing import Any, Iterator

import ndjson
from shapely.geometry import shape

from schematools.types import DatasetFieldSchema, DatasetTableSchema

from .base import BaseImporter, Provenance, Record


class NDJSONImporter(BaseImporter):
    """Import an NDJSON file into the database."""

    def parse_records(
        self, file_name: PosixPath, dataset_table: DatasetTableSchema, **kwargs: Any
    ) -> Iterator[dict[str, list[Record]]]:
        """Provide an iterator the reads the NDJSON records."""
        # Initializes the field mapper once for the table
        field_mapper = TableFieldMapper(dataset_table)
        with open(file_name) as fh:
            for row in ndjson.reader(fh):
                records = field_mapper.parse_object(row)
                yield records


class TableFieldMapper:
    """Conversion of field names during the import.
    The :meth:`parse_object` returns the database records to create for a given NDJson line.
    """

    # This is mapping the GOBModel `entity_id` to `Id`
    # that needs to be used in GraphQL queries.
    id_name_mapping = {
        "identificatie": "Id",
        "volgnummer": "Volgnummer",
        "dossier": "Id",
        "documentnummer": "Id",
        "vestigingsnummer": "Id",
        "sbiActiviteitNummer": "Id",
        "kvknummer": "Id",
        "wozdeelobjectnummer": "Id",
        "wozobjectnummer": "Id",
        "id": "Id",
        "code": "Id",
        "neuronId": "Id",
    }

    def __init__(self, dataset_table: DatasetTableSchema):
        """Analysis of fields that need special attention during the import."""
        self.dataset_table = dataset_table

        # Create cached provenance objects for all fields that have it.
        # The fields don't return a Provenance object themselves,
        # so provenance can also point to completely different datasource in the future.
        self.cached_provenance: dict[str, Provenance] = {
            prov: Provenance(prov) for field in dataset_table.fields if (prov := field.provenance)
        }

        # XXX maybe this is too much of a dirty hack, and it would be better
        # to have an external configuration that determines which fields should
        # be flattened to strings
        self.inactive_relation_info = {
            field.id
            for field in dataset_table.fields
            if (comment := field.get("$comment")) and "*stringify*" in comment
        }

        # When the table is a through-table, the GOB model fields need to be fetched elsewhere.
        self.through_field_map = {}
        if dataset_table.is_through_table:
            for prefix, through_field in zip(("src", "dst"), dataset_table.through_fields):
                self.through_field_map[through_field.id] = [
                    # generate "srcId", "dstId" and "srcVolgnummer", "dstVolgnummer" fields.
                    (related_identifier, f"{prefix}{self.id_name_mapping[related_identifier]}")
                    for related_identifier in through_field.related_table.identifier
                ]

    def parse_object(self, source: dict) -> dict[str, list[Record]]:
        """Parse the record, convert field names.
        Returns all database records to create, grouped by SQL table name.
        """
        main_row = Record(data={}, source=source)
        sub_rows = {}

        self._fix_through_fields(source)
        composite_key_filled = self._fill_composite_pk(source, main_row)

        # Fill all standard fields from the source.
        for field in self.dataset_table.get_fields(include_subfields=True):
            if field.id == "schema" and field.type.startswith("https://"):
                continue
            if field.id == "id" and composite_key_filled:
                # The composite key is already inserted in main_row
                # skip processing "id" field again
                continue

            # Read the value from the record. As subfields are also part of the main loop,
            # extra care is needed to make sure the correct dictionary is checked for the value.
            try:
                field_source = source[field.parent_field.id] if field.is_subfield else source
                value = self._get_value(field, field_source)
            except LookupError:
                if field.is_identifier_part:
                    # Avoid overriding the identifier we've just generated with None.
                    continue
                # Some missing fields still need to be mentioned in the insert statement.
                value = None

            if field.is_nested_table:
                # Nested object
                if value:
                    sub_rows[field.nested_table.id] = self._format_nested_rows(
                        field, value, main_row
                    )
            elif field.nm_relation is not None:
                # M2M through table
                if value:
                    sub_rows[field.through_table.id] = self._format_through_rows(
                        field, value, source
                    )
            elif field.relation is not None and field.is_object and isinstance(value, dict):
                # Foreign key to temporal (composite) key
                self._fill_composite_fk(field, value, main_row)
            else:
                # Any other field.
                main_row[field.db_name] = self._format_db_value(field, value)

        return {
            self.dataset_table.id: [main_row],
            **sub_rows,
        }

    def _fill_composite_pk(self, source: dict, row: Record) -> bool:
        """Adds a composite key 'id' for temporal tables.
        Return value indicates whether this was done.
        """
        temporal = not self.dataset_table.is_autoincrement and self.dataset_table.has_composite_key
        if not temporal:
            return False

        row["id"] = self._get_composite_id(source)
        return True

    def _get_value(self, field: DatasetFieldSchema, source: dict) -> Any:
        """Retrieve a single value from the source object.
        Raises LookupError when the value does not exist in the source.
        Note that any subfields are not treated differently,
        its assumed the 'source' is already at the correct level.
        """
        if source is None and field.is_subfield:
            raise LookupError(field.id)

        if field.provenance:
            # JSONPath or alias.
            return self.cached_provenance[field.provenance].resolve(source)
        else:
            # camelCase name
            return source[field.id]

    def _format_db_value(self, field: DatasetFieldSchema, value) -> Any:
        """Adjust the value for database format."""
        if value is None:
            return None
        elif field.is_geo:
            # Format geometry fields
            wkt = shape(value).wkt
            return f"SRID={field.srid};{wkt}"
        elif field.id in self.inactive_relation_info:
            # Convert nested object to JSON string
            return json.dumps(value)
        elif isinstance(value, (dict, list)):
            raise ValueError(
                f"Value of '{field.qualified_id}' should resolve to a scalar, not: {value!r}"
            )
        else:
            return value

    def _get_composite_id(self, row: dict) -> str:
        """Concat identifier fields for a single composite field value"""
        return ".".join(str(row[fn]) for fn in self.dataset_table.identifier)

    def _fix_through_fields(self, source: dict):
        """Maps fields in ndjson to proper fieldnames and structure.

        When through tables (1-N and NM relations) are imported directly
        from the GOB GraphQL API, the fieldnames in the ndjson are not
        following the names of the associated Amsterdam Schema.

        The importer is built around the Amsterdam Schema fieldnames.
        So, incoming ndjson records needs to be mapped to the appropriate
        amsterdam schema field names, to be able to import the ndjson
        records correctly.
        """
        for through_field_id, gob_names in self.through_field_map.items():
            if len(gob_names) > 1:  # composite key
                source[through_field_id] = {
                    identifier: source[gob_name] for identifier, gob_name in gob_names
                }
            else:
                gob_name = gob_names[0][1]
                source[through_field_id] = source[gob_name]

    def _fill_composite_fk(self, rel_field: DatasetFieldSchema, value: dict | None, row: Record):
        """Process a composite foreign key"""
        # Flatten subfields, as they are part of the same record.
        fk_value_parts = []
        for subfield in rel_field.subfields:
            # Any date-ranges of a composite key are not inlined in the main object
            if subfield.is_temporal_range:
                continue

            if value is None:
                subfield_value = None
            else:
                subfield_value = value[subfield.id]
                fk_value_parts.append(subfield_value)

            row[subfield.db_name] = subfield_value

        # Make sure composite keys also get their Django _id field.
        # Empty fk_value_parts should result in None value
        row[rel_field.db_name] = ".".join([str(p) for p in fk_value_parts]) or None

    def _format_nested_rows(
        self, n_field: DatasetFieldSchema, value: list[dict], main_row: Record
    ) -> list[Record]:
        """Collect the records for a nested field"""
        # When the identifier is composite, we can assume
        # that an extra 'id' field will be available, because
        # Django cannot live without it.
        parent_id_fields = self.dataset_table.identifier
        parent_id_field = "id" if len(parent_id_fields) > 1 else parent_id_fields[0]
        parent_id = main_row[parent_id_field]

        return [
            Record(
                {
                    "parent_id": parent_id,
                    **self._get_field_values(n_field.subfields, nested_row),
                },
                source=nested_row,
            )
            for nested_row in value
        ]

    def _get_field_values(self, fields: Iterator[DatasetFieldSchema], nested_row: dict):
        """Extract any (nested) values from a row."""
        row = {}
        for field in fields:
            try:
                row[field.db_name] = self._get_value(field, nested_row)
            except LookupError:
                pass  # skip unmentioned fields

        return row

    def _format_through_rows(
        self, nm_field: DatasetFieldSchema, values: list[dict], source: dict
    ) -> list[Record]:
        """Provide the records for a Many-to-Many (NM) fields."""
        source_db_name = self.dataset_table.db_name_variant(with_dataset_prefix=False)
        src_id_value = self._get_composite_id(source)
        if not isinstance(values, list):
            values = [values]

        through_row_records = []
        for value in values:  # type: dict[str, Any]
            # Generate foreignkey value for right hand side
            dst_id_value = ".".join(
                str(value[field.id]) for field in nm_field.subfields if not field.is_temporal_range
            )

            # Generate the main id values that Django needs
            through_row_record = Record(
                {
                    f"{source_db_name}_id": src_id_value,
                    f"{nm_field.db_name}_id": dst_id_value,
                },
                source=value,
            )

            # Fill the temporal identifier fields if these are present.
            for prefix, table, source_dict in (
                (source_db_name, self.dataset_table, source),
                (nm_field.db_name, nm_field.related_table, value),
            ):
                if table.has_composite_key:
                    for id_field in table.identifier_fields:
                        subfield_id = f"{prefix}_{id_field.db_name}"
                        through_row_record[subfield_id] = source_dict[id_field.id]

            # Fill any temporal dimensions if these are part of the through record
            for subfield in nm_field.subfields:
                if subfield.is_temporal_range:
                    try:
                        through_row_record[subfield.db_name] = self._get_value(subfield, value)
                    except KeyError:
                        pass

            through_row_records.append(through_row_record)

        return through_row_records



"""This module contains converters to the CKAN format.

CKAN (https://ckan.org) is the system used by the metadata catalog at
https://data.overheid.nl.
"""

from ._convert import from_dataset

__all__ = ["from_dataset"]



import json

from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _


@deconstructible
class URLPathValidator(RegexValidator):
    regex = r"\A[a-z0-9]+([/-][a-z0-9]+)*\Z"
    message = _("Only these characters are allowed: a-z, 0-9, '-' and '/' between paths.")


def validate_json(value: str) -> None:
    """Validate the contents of a JSON value."""
    try:
        json.loads(value)
    except (TypeError, ValueError):
        raise ValidationError(
            _("Value must be valid JSON text."),
            code="invalid",
            params={"value": value},
        )



from dataclasses import asdict, dataclass

import mappyfile

#  Serializing with Jinja templates
from jinja2 import Environment, FileSystemLoader

from . import types


def schema_dict(items):
    return {k: v for k, v in items if v is not None}


@dataclass
class MappyfileSerializer:
    template: str = "MAP END"

    def map_file_to_dict(self, map_file):
        return asdict(map_file, dict_factory=schema_dict)

    def __call__(self, map_file: types.Mapfile):
        backend_file = mappyfile.loads(self.template)
        backend_file.update(self.map_file_to_dict(map_file))
        errors = mappyfile.validate(backend_file)
        if errors:
            raise Exception(errors)
        return mappyfile.dumps(backend_file)


@dataclass
class JinjaSerializer:
    template_dir: str

    @property
    def env(self) -> Environment:
        env = Environment(loader=FileSystemLoader(self.template_dir))  # nosec
        env.trim_blocks = True
        env.lstrip_blocks = True
        return env

    def __call__(self, template_file_name: str, context: dict):
        template = self.env.get_template(template_file_name)
        return template.render(**context)



from typing import Any, Dict, List

from django.core.management import CommandParser

from schematools.contrib.django.faker.create import create_data_for

from . import BaseDatasetCommand


class Command(BaseDatasetCommand):  # noqa: D101
    help = """Create mock data for Amsterdam schema files.

    Datasets (in DSO db) + dataset tables should already have been created,
    usually with the `import_schemas --create-tables` mgm. command.
    """  # noqa: A003
    requires_system_checks = []

    def add_arguments(self, parser: CommandParser) -> None:  # noqa: D102
        super().add_arguments(parser)
        parser.add_argument("-s", "--size", type=int, default=50, help="Number of rows")
        parser.add_argument("--sql", action="store_true", help="Generate the sql statements.")
        parser.add_argument(
            "--start-at", type=int, default=1, help="Starting number for sequences."
        )

    def handle(self, *args: List[Any], **options: Dict[str, Any]) -> None:  # noqa: D102
        datasets = self.get_datasets(options, enable_db=True, default_all=False)
        sql_lines = create_data_for(
            *datasets,
            start_at=options["start_at"],
            size=options["size"],
            sql=options["sql"],
        )
        if sql_lines:
            self.stdout.write("\n".join(sql_lines))



import io

from django.core.management import BaseCommand, CommandError
from django.db import connection
from django.db.utils import ProgrammingError
from psycopg2 import sql

from schematools.contrib.django.models import Dataset, DatasetTable
from schematools.naming import to_snake_case


class Command(BaseCommand):
    """Remove Dataset objects and optionally their underlying tables."""

    help = "Drop the specified schemas"  # noqa: A003
    requires_system_checks = []

    def add_arguments(self, parser):
        """Django hook."""
        parser.add_argument(
            "schemas",
            nargs="+",
            help="Schemas that need to be dropped",
        )
        parser.add_argument(
            "--drop-tables",
            action="store_true",
            default=False,
            help="Also drop the tables associated with the schemas",
        )

    def handle(self, *args, **options):
        """Django hook implementing command logic."""
        datasets = Dataset.objects.all()
        imported_datasets = {d.name for d in datasets}
        drop_schemas = set(options.get("schemas", []))

        impossible_schemas = [ident for ident in drop_schemas if ident not in imported_datasets]
        if impossible_schemas:
            msg = io.StringIO()
            msg.write("These schemas do not exist:\n")

            for ident in impossible_schemas:
                msg.write(f"* {ident!r}")
                snake = to_snake_case(ident)
                if snake in imported_datasets:
                    msg.write(f" (did you mean {snake!r}?)")
                msg.write("\n")

            raise CommandError(msg.getvalue())

        dataset_qs = datasets.filter(name__in=drop_schemas)
        if options["drop_tables"]:
            tables = DatasetTable.objects.filter(dataset__in=dataset_qs)
            with connection.cursor() as cursor:
                for table in tables:
                    name = table.db_table
                    try:
                        cursor.execute(
                            sql.SQL("DROP TABLE {table} CASCADE;").format(
                                table=sql.Identifier(name)
                            )
                        )
                        self.stdout.write(f"Deleted table {name}")
                    except ProgrammingError as e:
                        # Dataset(Table) descriptions and their generated tables
                        # exist independently. It is therefore possible
                        # that the table is already deleted or does not exist.
                        # We only need to attempt to clean it here.
                        self.stdout.write(f"Failed to delete table {name}. Error: {e}")
                        continue

        dataset_qs.delete()
        if options["verbosity"] > 0:
            self.stdout.write(f"Deleted the following datasets: {drop_schemas}")



import json
import re
from typing import Any, Iterable, Tuple

from shapely.geometry import shape

from schematools.exceptions import ParserError
from schematools.types import DatasetTableSchema

from .base import BaseImporter, Record

ID_FORMAT = re.compile(r"^([a-z0-9_]+)[/.](\d+)$", re.I)


def read_geojson(file_name) -> Iterable[dict]:
    """Read the contents of a GeoJSON file into memory.
    This also performs basic checks whether the data is valid.
    """
    with open(file_name) as fh:
        geojson = json.load(fh)
        if (
            not isinstance(geojson, dict)
            or geojson.get("type") != "FeatureCollection"
            or not isinstance(geojson.get("features"), list)
        ):
            raise ParserError(f"{file_name} is not a valid GeoJSON file")

    for feature in geojson["features"]:
        feature_type = feature.get("type")
        if feature_type != "Feature":
            raise ParserError(f"Expected 'Feature' in {file_name}, not {feature_type}")

        yield feature


def split_id(id_value) -> Tuple[str, str]:
    # When the ID format is name/identifier,
    # this detects that different feature types are part of the same file.
    match = ID_FORMAT.match(id_value)
    if match:
        return match.group(1), match.group(2)
    else:
        raise ValueError("Can't split ID value")


class GeoJSONImporter(BaseImporter):
    """Import an GeoJSON file into the database."""

    def parse_records(self, file_name, dataset_table: DatasetTableSchema, **kwargs):
        """Provide an iterator the reads the NDJSON records"""
        features = read_geojson(file_name)
        main_geometry = dataset_table.main_geometry_field

        relation_fields = [field for field in dataset_table.fields if field.relation is not None]
        for feature in features:
            record = Record(
                (self._clean_value(name, value) for name, value in feature["properties"].items()),
                source=feature,
            )
            for relation_field in relation_fields:
                if relation_field.db_name != relation_field.python_name:
                    record[relation_field.db_name] = record.pop(relation_field.python_name)

            wkt = shape(feature["geometry"]).wkt
            record[main_geometry.db_name] = f"SRID={main_geometry.srid};{wkt}"

            yield {dataset_table.id: [record]}

    def _clean_value(self, name: str, value: Any) -> Tuple[str, Any]:
        if name[0] in "@$":
            name = name[1:]

        if name == "id":
            try:
                value = split_id(value)[1]
            except ValueError:
                pass

        return name, value



import re
from time import strftime
from typing import Any

from ..types import DatasetSchema


def from_dataset(ds: DatasetSchema, path: str) -> dict[str, Any]:
    """Convert a dataset to the CKAN format variant used by data.overheid.nl.

    path should be the dataset path as used in DSO-API, e.g., "beheerkaart/cbs_grid"
    for the dataset with id "beheerkaartCbsGrid".

    The output, as JSON, can be used in a CKAN create_package request.

    A limited amount of documentation for the format can be pieced together from:
    * https://docs.ckan.org/en/2.9/api/index.html
    * https://ckanext-dcatdonl.readthedocs.io/en/latest/usage-creationandupdating.html
    * https://waardelijsten.dcat-ap-donl.nl/
    """

    # XXX CKAN allows a dataset to have multiple resources associated with it,
    # which we could use to represent the tables, but data.overheid.nl doesn't
    # seem to support that (or it's broken).

    url = "https://api.data.amsterdam.nl/v1/" + path

    has_geo = any(table.has_geometry_fields for table in ds.tables)
    theme = THEME_RUIMTE if has_geo else THEME_BESTUUR

    # Apparently, titles must be globally unique. If they're not, we get an error message that
    # complains about non-unique identifiers. So, add "_amsterdam_dso-api" to prevent clashes.
    #
    # TODO if we upload anything with uppercase letters or spaces, we get an error message.
    # But most datasets in the catalog at data.overheid.nl do use those. Figure out how to
    # preserve uppercase and spaces.
    title = (
        ds.title.lower().replace(" ", "_")
        if ds.title is not None and ALMOST_NAME.match(ds.title)
        else ds.id.lower()
    ) + "_amsterdam_dso-api"

    data = {
        "authority": "http://standaarden.overheid.nl/owms/terms/Amsterdam",
        "contact_point_name": "Gemeente Amsterdam",
        "contact_point_email": "datapunt@amsterdam.nl",
        "contact_point_website": "https://data.amsterdam.nl",
        "format": JSON,
        "identifier": url,
        "language": [NLD],
        "license_id": LICENSES.get(ds.license, LICENSE_UNKNOWN),
        "metadata_language": NLD,
        "mimetype": "application/json",
        "modified": strftime("%Y-%m-%dT%H:%M:%S"),
        "name": title,
        "notes": ds.description or title,  # Must not be empty.
        "owner_org": "gemeente-amsterdam",
        "publisher": "http://standaarden.overheid.nl/owms/terms/Amsterdam",
        "theme": [theme],
        "title": title,
        "url": url,
    }

    # XXX I can't find what the inverse is called, so we omit this field for
    # niet_beschikbaar. The uploader in dcatd does the same.
    if ds.status == DatasetSchema.Status.beschikbaar:
        data["dataset_status"] = "http://data.overheid.nl/status/beschikbaar"

    return data


# RE that describes valid CKAN names and titles, except that it also allows lowercase and space.
# We map the latter to _.
ALMOST_NAME = re.compile(r"^[A-Za-z0-9_ -]+$")

# JSON is our primary file type. Not sure if we can submit multiple types for a dataset.
# Keep this consistent with the mimetype.
# From https://waardelijsten.dcat-ap-donl.nl/mdr_filetype_nal.json
JSON = "http://publications.europa.eu/resource/authority/file-type/JSON"

# From https://waardelijsten.dcat-ap-donl.nl/overheid_license.json
LICENSES = {
    "Creative Commons, Naamsvermelding": "http://creativecommons.org/licenses/by/4.0/deed.nl",
    "public": "http://creativecommons.org/publicdomain/mark/1.0/deed.nl",
    "Gesloten": "http://standaarden.overheid.nl/owms/terms/geslotenlicentie",
}
LICENSE_UNKNOWN = "http://standaarden.overheid.nl/owms/terms/licentieonbekend"

# From https://waardelijsten.dcat-ap-donl.nl/donl_language.json
NLD = "http://publications.europa.eu/resource/authority/language/NLD"

# From
# https://github.com/dataoverheid/donlsync-mappings/blob/1c7b23f3f6dfc623/value-mappings/Eindhoven__Dataset__AccessRights.json
THEME_BESTUUR = "http://standaarden.overheid.nl/owms/terms/Bestuur"
THEME_RUIMTE = "http://standaarden.overheid.nl/owms/terms/Ruimte_en_infrastructuur"



"""Exporter module.

This module facilitates the export of events from GOB database tables.
The main objective of this module is to create full events that can be used
to test the event-importer that is also part of the `schematools.events` package.

The `schematools` cli interface had an entry to use the code in this module.
"""
from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, Iterator

from geoalchemy2.shape import to_shape
from json_encoder import json
from sqlalchemy import Table
from sqlalchemy.engine import Connection

from schematools.events import metadata
from schematools.factories import tables_factory
from schematools.naming import to_snake_case
from schematools.types import DatasetSchema, DatasetTableSchema


@dataclass
class ComplexFieldAttrs:
    field_name: str
    is_many: bool | None = None
    relation_ds: str | None = None
    identifier_names: list[str] = field(default_factory=list)
    subfield_names: list[str] = field(default_factory=list)


def fetch_complex_fields_metadata(dataset_table: DatasetTableSchema) -> list[ComplexFieldAttrs]:
    """Collect info about complex fields (mainly relations)."""
    complex_fields_metadata = []

    for complex_field in dataset_table.complex_fields:
        is_many = None
        properties = {}
        relation_ds = None
        try:
            if (nm_relation := complex_field.nm_relation) is not None:
                is_many = True
                properties = complex_field["items"]["properties"]
                relation_ds = nm_relation.split(":")[0]
            if (relation := complex_field.relation) is not None:
                is_many = False
                properties = complex_field["properties"]
                relation_ds = relation.split(":")[0]
        except KeyError:
            # Simple (non-object) relations do not have "properties" nor "items".
            # Those relations are not relevant for the export,
            # so they will be skipped.
            continue

        if is_many is not None:
            complex_fields_metadata.append(
                ComplexFieldAttrs(
                    complex_field_name=to_snake_case(complex_field.name),
                    is_many=is_many,
                    relation_ds=relation_ds,
                    identifier_names=dataset_table.identifier,
                    subfield_names=[to_snake_case(sf) for sf in properties.keys()],
                )
            )

    return complex_fields_metadata


def collect_nm_embed_rows(
    dataset_id: str,
    table_id: str,
    datasets_lookup: dict[str, DatasetSchema],
    tables: dict[str, dict[str, Table]],
    complex_fields_metadata: list[ComplexFieldAttrs],
    connection: Connection,
) -> dict[str, Any]:
    """Fetch row info as list of embeddable objects."""
    nm_embeds = defaultdict(lambda: defaultdict(list))
    for field_metadata in complex_fields_metadata:
        if field_metadata.is_many:
            through_table = tables[dataset_id][f"{table_id}_{field_metadata.field_name}"]
            for row in connection.execute(through_table.select()):
                row_dict = dict(row)
                id_value = ".".join(
                    str(row_dict[f"{table_id}_{idn}"]) for idn in field_metadata.identifier_names
                )

                stripped_row = {}
                for sfn in field_metadata.subfield_names:
                    stripped_row[sfn] = row_dict[f"{field_metadata.field_name}_{sfn}"]
                nm_embeds[table_id][id_value].append(stripped_row)
    return dict(nm_embeds)


def fetch_nm_embeds(
    row: dict[str, Any],
    table_id: str,
    nm_embed_rows: dict[str, Any],
    complex_fields_metadata: list[ComplexFieldAttrs],
) -> dict[str, list[dict[str, Any]]]:
    """Fetch row info as lists of embeddables."""
    nm_embeds = defaultdict(list)
    for field_metadata in complex_fields_metadata:
        if field_metadata.is_many:
            id_value = ".".join(str(row[idn]) for idn in field_metadata.identifier_names)
            row_dicts = nm_embed_rows[table_id].get(id_value)
            nm_embeds[field_metadata.field_name] = row_dicts

    return dict(nm_embeds)


def fetch_1n_embeds(
    row: dict[str, Any], complex_fields_metadata: list[ComplexFieldAttrs]
) -> dict[str, dict[str, Any]]:
    """Fetch row info as embeddable object(s)."""
    embeddable_objs = {}
    for field_metadata in complex_fields_metadata:
        if not field_metadata.is_many:
            embed_obj = {}
            for subfield_name in field_metadata.subfield_names:
                embed_obj[subfield_name] = row[f"{field_metadata.field_name}_{subfield_name}"]
            embeddable_objs[field_metadata.field_name] = embed_obj
    return embeddable_objs


def export_events(
    datasets, dataset_id: str, table_id: str, connection: Connection
) -> Iterator[str]:
    """Export the events from the indicated dataset and table."""
    tables: dict[str, dict[str, Table]] = {}
    datasets_lookup: dict[str, DatasetSchema] = {ds.id: ds for ds in datasets}
    dataset_table: DatasetTableSchema = datasets_lookup[dataset_id].get_table_by_id(table_id)
    geo_fields = [field.python_name for field in dataset_table.fields if field.is_geo]

    complex_fields_metadata: list[ComplexFieldAttrs] = fetch_complex_fields_metadata(dataset_table)

    for ds_id, dataset in datasets_lookup.items():
        tables[ds_id] = tables_factory(dataset, metadata)

    # Collect in one go (to prevent multiple queries)
    nm_embed_rows: dict[str, Any] = collect_nm_embed_rows(
        dataset_id, table_id, datasets_lookup, tables, complex_fields_metadata, connection
    )
    for r in connection.execute(tables[dataset_id][table_id].select()):
        row = dict(r)
        meta = {"event_type": "ADD", "dataset_id": dataset_id, "table_id": table_id}
        id_ = ".".join(str(row[f]) for f in dataset_table.identifier)
        event_parts = [f"{dataset_id}.{table_id}.{id_}", json.dumps(meta)]
        for geo_field in geo_fields:
            geom = row.get(geo_field)
            if geom:
                row[geo_field] = f"SRID={geom.srid};{to_shape(geom).wkt}"
        row.update(fetch_1n_embeds(row, complex_fields_metadata))
        row.update(fetch_nm_embeds(row, table_id, nm_embed_rows, complex_fields_metadata))
        event_parts.append(json.dumps(row))
        yield "|".join(event_parts)



"""Create GRANT statements to give roles very specific access to the database."""
from __future__ import annotations

import logging
from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, cast

from pg_grant import PgObjectType, parse_acl_item, query
from pg_grant.sql import grant, revoke
from sqlalchemy import event, text
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session, sessionmaker

from schematools.permissions import PUBLIC_SCOPE  # type: ignore [attr-defined]
from schematools.types import DatasetSchema

# Create a module-level logger, so calling code can
# configure the logger, if needed.
logger = logging.getLogger(__name__)

existing_roles = set()


def is_remote(table_name: str) -> bool:
    """Test if table_name refers a remote table.

    WARNING: UGLY HACK until 265311 is resolved and we
    can interrogate schematools to find out whether a table
    is remote.
    """
    return (
        table_name.startswith("haalcentraalbrk")
        or table_name.startswith("haalcentraalbag")
        or table_name.startswith("brp")
    )


def introspect_permissions(engine: Engine, role: str) -> None:
    """Shows the table permissions."""
    schema_relation_infolist = query.get_all_table_acls(engine, schema="public")
    for schema_relation_info in schema_relation_infolist:
        if schema_relation_info.acl:
            acl_list = [parse_acl_item(item) for item in schema_relation_info.acl]
            for acl in acl_list:
                if acl.grantee == role:
                    logger.info(
                        'role "%s" has privileges %s on table "%s"',
                        role,
                        ",".join(acl.privs),
                        schema_relation_info.name,
                    )


def revoke_permissions(engine: Engine, role: str, verbose: int = 0) -> None:
    """Revoke all privileges for the indicated role."""
    grantee = role
    schema_relation_infolist = query.get_all_table_acls(engine, schema="public")
    for schema_relation_info in schema_relation_infolist:
        if schema_relation_info.acl:
            acl_list = [parse_acl_item(item) for item in schema_relation_info.acl]
            for acl in acl_list:
                if acl.grantee == role:
                    if verbose:
                        logger.info(
                            'revoking ALL privileges of role "%s" on table "%s"',
                            role,
                            schema_relation_info.name,
                        )
                    revoke_statement = revoke(
                        "ALL", PgObjectType.TABLE, schema_relation_info.name, grantee
                    )
                    engine.execute(revoke_statement)


def apply_schema_and_profile_permissions(
    engine: Engine,
    pg_schema: str,
    ams_schema: DatasetSchema | dict[str, DatasetSchema],
    profiles: dict[str, Any],
    role: str,
    scope: str,
    set_read_permissions: bool = True,
    set_write_permissions: bool = True,
    dry_run: bool = False,
    create_roles: bool = False,
    revoke: bool = False,
    verbose: int = 0,
) -> None:
    """Apply permissions for schema and profile."""
    SessionCls = sessionmaker(bind=engine)
    session = SessionCls()

    if verbose:

        @event.listens_for(engine, "after_cursor_execute")
        def after_cursor_execute(conn, cursor, statement, parameters, context, executemany):
            for notice in cursor.connection.notices:
                logger.info(notice.strip())

    try:
        if ams_schema:
            create_acl_from_schemas(
                session,
                pg_schema,
                ams_schema,
                role,
                scope,
                set_read_permissions,
                set_write_permissions,
                dry_run,
                create_roles,
                revoke,
                verbose,
            )
        if profiles:
            profile_list = cast(List[Dict[str, Any]], profiles.values())
            create_acl_from_profiles(engine, pg_schema, profile_list, role, scope, verbose)
        session.commit()
    except Exception:
        session.rollback()
        logger.warning("Session rolled back")
        raise
    finally:
        session.close()


def create_acl_from_profiles(
    engine: Engine,
    pg_schema: str,
    profile_list: list[dict[str, Any]],
    role: str,
    scope: str,
    verbose: int = 0,
) -> None:
    """Create an ACL from profile list.

    NOTE: Rudimentary, not ready for production!
    """
    acl_list = query.get_all_table_acls(engine, schema=pg_schema)
    privileges = [
        "SELECT",
    ]
    grantee = role
    for profile in profile_list:
        if scope in profile["scopes"]:
            for dataset, _details in profile["schema_data"]["datasets"].items():
                for item in acl_list:
                    if item.name.startswith(dataset + "_"):
                        grant_statement = grant(
                            privileges,
                            PgObjectType.TABLE,
                            item.name,
                            grantee,
                            grant_option=False,
                            schema=pg_schema,
                        )
                        if verbose:
                            logger.info(grant_statement)
                        engine.execute(grant_statement)


def set_dataset_write_permissions(
    session: Session,
    pg_schema: str,
    ams_schema: DatasetSchema,
    dry_run: bool,
    create_roles: bool,
    echo: bool = False,
) -> None:
    """Sets write permissions for the indicated dataset."""
    grantee = f"write_{ams_schema.db_name}"
    if create_roles:
        _create_role_if_not_exists(session, grantee, dry_run=dry_run)
    for table in ams_schema.get_tables(include_nested=True, include_through=True):
        table_name = table.db_name
        if is_remote(table_name):
            continue
        table_privileges = ["SELECT", "INSERT", "UPDATE", "DELETE", "TRUNCATE", "REFERENCES"]
        _execute_grant(
            session,
            grant(
                table_privileges,
                PgObjectType.TABLE,
                table_name,
                grantee,
                grant_option=False,
                schema=pg_schema,
            ),
            echo=echo,
            dry_run=dry_run,
        )


def get_all_dataset_scopes(
    ams_schema: DatasetSchema,
    role: str,
    scope: str,
) -> DefaultDict[str, list]:
    """Returns all scopes that should be applied to the tables of a dataset.

    Args:
        ams_schema: the amsterdam schema that needs to be processed
        role: the role that needs the grants that are calculated from the schema
            A special value `AUTO` can be used to apply grants for all scopes
        scope: only return grants for a specific scope, value can be empty string ("")

        The grants will be applied according to the configuration of the auth scopes
        in the amsterdam schema.
        If not auth scopes are defined, all tables get the `scope_openbaar` grant.
        If only the dataset has as scope `foo`, alle tables get the `scope_foo` grant.
        If a table has as scope `bar`, this overrules the dataset scope,
        so this table gets the `scope_bar` grant.
        If one or more fields have a scope, this scope overrules both the
        dataset and the table scope. The other fields in the same table
        get a `scope_openbaar` grant in that case.
        If a 1-N relation field has a scope, the foreign key field get the associated
        grant. In case this relation field is of type `object`` (e.g. for temporal fields),
        the additional columns (usually identificatie/volgnummer postfixed) get the same grant.
        If NM and nested relation fields (type `array` in the schema) have a scope `bar`
        the associated sub-table gets the grant `scope_bar`.

    Returns:
        all_scopes (defaultdict): Contains for each table in the dataset a list of scopes with
            priviliges and grants:

            '"table1":[
                        {
                            "privileges": ["SELECT"],
                            "grantees": ["scope_openbaar"]),
                        }
                    ],
            "table2":
                    [
                        {
                            "privileges": ["SELECT columnA"],
                            "grantees": ["scope_openbaar"]),
                        },
                        {
                            "privileges": ["SELECT columnB"],
                            "grantees": ["scope_A", "scope_B"]),
                        }
                    ]
                    '
    """

    def _fetch_grantees(scopes: frozenset[str]) -> list[str]:
        if role == "AUTO":
            grantees = [scope_to_role(scope) for scope in scopes]
        elif scope in scopes:
            grantees = [role]
        else:
            grantees = []
        return grantees

    all_scopes = defaultdict(list)
    dataset_scopes = ams_schema.auth

    for table in ams_schema.get_tables(include_nested=True, include_through=True):
        table_name = table.db_name
        if is_remote(table_name):
            continue

        table_scopes = table.auth
        fallback_scope = (table_scopes - {PUBLIC_SCOPE}) or dataset_scopes
        fields = [
            field for field in table.get_fields(include_subfields=True) if field.name != "schema"
        ]

        column_scopes = {}

        # First process all fields, to know if any fields has a non-public scope
        for field in fields:
            column_name = field.db_name
            # Object type relations have subfields, in that case
            # the auth scope on the relation is leading.
            parent_field_scopes: frozenset[str] = frozenset()
            if field.is_subfield:
                parent_field_scopes = field.parent_field.auth - {PUBLIC_SCOPE}

            field_scopes = field.auth - {PUBLIC_SCOPE}
            final_scopes: frozenset[str] = parent_field_scopes or field_scopes

            if final_scopes:
                column_scopes[column_name] = final_scopes

            if field.is_nested_table:
                all_scopes[field.nested_table.db_name].append(
                    {
                        "privileges": ["SELECT"],
                        "grantees": _fetch_grantees(final_scopes or fallback_scope),
                    }
                )

            if field.nm_relation is not None:
                all_scopes[field.through_table.db_name].append(
                    {
                        "privileges": ["SELECT"],
                        "grantees": _fetch_grantees(final_scopes or fallback_scope),
                    }
                )

        if column_scopes:
            for field in fields:
                if field.nm_relation or field.is_nested_table:
                    # field is not in view when nm or nested
                    continue

                column_name = field.db_name
                all_scopes[table_name].append(
                    # NB. space after SELECT is significant!
                    {
                        "privileges": [f"SELECT ({column_name})"],
                        "grantees": _fetch_grantees(
                            column_scopes.get(column_name, fallback_scope)
                        ),
                    }
                )
        else:
            if table_name not in all_scopes:
                all_scopes[table_name].append(
                    {
                        "privileges": ["SELECT"],
                        "grantees": _fetch_grantees(fallback_scope),
                    }
                )
    return all_scopes


def set_dataset_read_permissions(
    session: Session,
    pg_schema: str,
    ams_schema: DatasetSchema,
    role: str,
    scope: str,
    dry_run: bool,
    create_roles: bool,
    echo: bool = False,
) -> None:
    """Sets read permissions for the indicated dataset.

    Args:
        session: SQLAlchemy type session
        pg_schema: schema in the postgres database
        ams_schema: the amsterdam schema that needs to be processed
        role: the role that needs the grants that are calculated from the schema
            A special value `AUTO` can be used to apply grants for all scopes
        scope: only apply grants for a specific scope, value can be empty string ("")
        dry_run: do not apply the grants
        create_roles: boolean indicating that if certain roles are not in the postgres db,
            these roles need to be created.

        The grants will be applied according to the configuration of the auth scopes
        in the amsterdam schema.
        If not auth scopes are defined, all tables get the `scope_openbaar` grant.
        If only the dataset has as scope `foo`, alle tables get the `scope_foo` grant.
        If a table has as scope `bar`, this overrules the dataset scope,
        so this table gets the `scope_bar` grant.
        If one or more fields have a scope, this scope overrules both the
        dataset and the table scope. The other fields in the same table
        get a `scope_openbaar` grant in that case.
        If a 1-N relation field has a scope, the foreign key field get the associated
        grant. In case this relation field is of type `object`` (e.g. for temporal fields),
        the additional columns (usually identificatie/volgnummer postfixed) get the same grant.
        If NM and nested relation fields (type `array` in the schema) have a scope `bar`
        the associated sub-table gets the grant `scope_bar`.
    """
    grantee: str | None = f"write_{ams_schema.db_name}"

    grantee = None if role == "AUTO" else role
    if create_roles and grantee:
        _create_role_if_not_exists(session, grantee)

    all_scopes = get_all_dataset_scopes(ams_schema, role, scope)

    for table_name, grant_params in all_scopes.items():

        for grant_param in grant_params:
            for _grantee in grant_param["grantees"]:
                if create_roles:
                    _create_role_if_not_exists(session, _grantee, dry_run=dry_run)
                _execute_grant(
                    session,
                    grant(
                        grant_param["privileges"],
                        PgObjectType.TABLE,
                        table_name,
                        _grantee,
                        grant_option=False,
                        schema=pg_schema,
                    ),
                    echo=echo,
                    dry_run=dry_run,
                )


def create_acl_from_schemas(
    session: Session,
    pg_schema: str,
    schemas: DatasetSchema | dict[str, DatasetSchema],
    role: str,
    scope: str,
    set_read_permissions: bool,
    set_write_permissions: bool,
    dry_run: bool,
    create_roles: bool,
    revoke: bool,
    verbose: int = 0,
) -> None:
    """Create and set the ACL for automatically generated roles based on Amsterdam Schema.

    Read permissions are granted to roles 'scope_X', where X are scopes found in Amsterdam Schema
    Write permissions are granted to roles 'write_Y', where Y are dataset ids,
    for all tables belonging to the dataset.
    Revoke old privileges before assigning new in case new privileges are more restrictive.
    """
    if revoke:
        if role == "AUTO":
            if isinstance(schemas, DatasetSchema):
                # for a single dataset
                _revoke_all_privileges_from_read_and_write_roles(
                    session, pg_schema, schemas, dry_run=dry_run, echo=bool(verbose)
                )
            else:
                _revoke_all_privileges_from_read_and_write_roles(
                    session, pg_schema, dry_run=dry_run, echo=bool(verbose)
                )
        else:
            if isinstance(schemas, DatasetSchema):
                # for a single dataset
                _revoke_all_privileges_from_role(
                    session, pg_schema, role, schemas, dry_run=dry_run, echo=bool(verbose)
                )
            else:
                _revoke_all_privileges_from_role(
                    session, pg_schema, role, dry_run=dry_run, echo=bool(verbose)
                )

    if set_read_permissions:
        if isinstance(schemas, DatasetSchema):
            # for a single dataset
            set_dataset_read_permissions(
                session, pg_schema, schemas, role, scope, dry_run, create_roles, echo=bool(verbose)
            )
        else:
            for dataset_schema in schemas.values():
                set_dataset_read_permissions(
                    session,
                    pg_schema,
                    dataset_schema,
                    role,
                    scope,
                    dry_run,
                    create_roles,
                    echo=bool(verbose),
                )

    if set_write_permissions:
        if isinstance(schemas, DatasetSchema):
            # for a single dataset
            set_dataset_write_permissions(
                session, pg_schema, schemas, dry_run, create_roles, echo=bool(verbose)
            )
        else:
            for dataset_schema in schemas.values():
                set_dataset_write_permissions(
                    session, pg_schema, dataset_schema, dry_run, create_roles, echo=bool(verbose)
                )


def _revoke_all_privileges_from_role(
    session: Session,
    pg_schema: str,
    role: str,
    dataset: DatasetSchema | None = None,
    echo: bool = True,
    dry_run: bool = False,
) -> None:
    status_msg = "Skipped" if dry_run else "Executed"
    if dataset:
        # for a single dataset
        revoke_statements = []
        for table in dataset.tables:
            revoke_statements.append(
                f"REVOKE ALL PRIVILEGES ON {pg_schema}.{table.db_name} FROM {role}"
            )
        revoke_statement = ";".join(revoke_statements)
    else:
        revoke_statement = f"REVOKE ALL PRIVILEGES ON ALL TABLES IN {pg_schema} FROM {role}"
    sql_statement = f"""
        DO $$
        BEGIN
        f{revoke_statement};
        EXCEPTION
         WHEN undefined_object
         THEN RAISE NOTICE '%, skipping', SQLERRM USING ERRCODE = SQLSTATE;
        END
        $$
    """
    if echo:
        logger.info("%s --> %s", status_msg, revoke_statement)
    if not dry_run:
        session.execute(sql_statement)


def _revoke_all_privileges_from_read_and_write_roles(
    session: Session,
    pg_schema: str,
    dataset: DatasetSchema | None = None,
    echo: bool = True,
    dry_run: bool = False,
) -> None:
    """Revoke all privileges that may have been previously granted.

    This is about grants to the scope_* and write_* roles.
    If dataset is provided, revoke only rights to the tables belonging to
    dataset.
    """
    status_msg = "Skipped" if dry_run else "Executed"
    # with engine.begin() as connection:
    result = session.execute(
        text(
            r"""
            SELECT rolname
            FROM pg_roles
            WHERE rolname LIKE 'scope\_%'
               OR rolname LIKE 'write\_%'
            """
        )
    )
    for rolname in result:
        if dataset:
            # for a single dataset
            revoke_statements = []
            for table in dataset.tables:
                revoke_statements.append(
                    f"REVOKE ALL PRIVILEGES ON {pg_schema}.{table.db_name} FROM {rolname[0]}"
                )
            revoke_statement = ";".join(revoke_statements)
        else:
            revoke_statement = (
                f"REVOKE ALL PRIVILEGES ON ALL TABLES IN SCHEMA {pg_schema} FROM {rolname[0]}"
            )
        revoke_block_statement = f"""
            DO
            $$
                BEGIN
                    {revoke_statement};
                EXCEPTION
                    WHEN insufficient_privilege THEN
                        RAISE NOTICE '%, skipping', SQLERRM USING ERRCODE = SQLSTATE;
                END
            $$
        """

        if echo:
            logger.info("%s --> %s", status_msg, revoke_statement)
        if not dry_run:
            session.execute(revoke_block_statement)


def _execute_grant(
    session: Session, grant_statement: str, echo: bool = True, dry_run: bool = False
) -> None:
    status_msg = "Skipped" if dry_run else "Executed"
    sql_statement = f"""
        DO
        $$
            BEGIN
                {grant_statement};
            EXCEPTION
              WHEN undefined_table THEN
                RAISE NOTICE '%, skipping', SQLERRM USING ERRCODE = SQLSTATE;
            END
        $$
        """
    if echo:
        logger.info("%s --> %s", status_msg, grant_statement)
    if not dry_run:
        session.execute(sql_statement)


def _create_role_if_not_exists(
    session: Session, role: str, echo: bool = True, dry_run: bool = False
) -> None:
    """Wrap the create role statement in an anonymous code block.

    Reason is to be able to catch exceptions.
    Don't break out of the session just because the role already exists
    """
    status_msg = "Skipped" if dry_run else "Executed"
    create_role_statement = f"CREATE ROLE {role}"
    if role not in existing_roles:
        sql_statement = f"""
            DO $$
            BEGIN
            {create_role_statement};
            EXCEPTION
             WHEN duplicate_object
             THEN RAISE NOTICE '%, skipping', SQLERRM USING ERRCODE = SQLSTATE;
             WHEN undefined_object
             THEN RAISE NOTICE '%, skipping', SQLERRM USING ERRCODE = SQLSTATE;
            END
            $$
        """
        if echo:
            logger.info("%s --> %s", status_msg, create_role_statement)
        if not dry_run:
            session.execute(text(sql_statement))
        existing_roles.add(role)


def scope_to_role(scope: str) -> str:
    """Return rolename for the postgres database."""
    return f"scope_{scope.lower().replace('/', '_')}"



"""Kafka consumer."""
from __future__ import annotations

import json
import logging
import os
from typing import Iterable

from confluent_kafka import Consumer, Message
from sqlalchemy.engine import Connection

from schematools.events.full import EventsProcessor
from schematools.types import DatasetSchema

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def _to_env_name(name: str) -> str:
    """Convert param name to env var name."""
    return name.replace(".", "_").upper()


def _fetch_consumer_params() -> dict:
    """Create the parameters for the consumer.

    A parameter is only included, if the associated env. var is available.
    """
    params = {}
    for name in (
        "bootstrap.servers",
        "sasl.mechanisms",
        "security.protocol",
        "sasl.username",
        "sasl.password",
        "group.id",
        "auto.offset.reset",
    ):
        if (env_name := _to_env_name(name)) is not None:
            if env_name in os.environ:
                params[name] = os.environ[env_name]
    return params


def consume_events(
    dataset_schemas: list[DatasetSchema],
    connection: Connection,
    topics: Iterable[str],
    truncate: bool = False,
):
    """Consume events.

    Main entry point for this module.
    """
    logger.debug("Starting script..")

    # Create Consumer instance 'auto.offset.reset=earliest'
    # to start reading from the beginning of the topic if no committed offsets exist
    consumer = Consumer(_fetch_consumer_params())
    # consumer = Consumer(
    #     {
    #         "bootstrap.servers": BOOTSTRAP_SERVERS,
    #         "auto.offset.reset": AUTO_OFFSET_RESET,
    #         "group.id": "test",
    #     }
    # )

    logger.debug("Consumer made")
    # Transform and Load

    importer = EventsProcessor(dataset_schemas, connection, truncate=truncate)
    logger.debug("Created importer")
    # Subscribe to topic, confluent_kafka explictly needs a list
    consumer.subscribe(list(topics))
    logger.debug("Starting consuming..")
    # Process messages
    try:
        while True:
            msg = consumer.poll(10.0)
            if msg is None:
                # No message available within timeout.
                # Initial message consumption may take up to `session.timeout.ms`
                # for the consumer group to rebalance and start consuming
                logger.debug("Waiting for message or event/error in poll()")
                continue
            elif msg.error():
                logger.error("error: %s", msg.error())
            else:
                process_message(importer, msg)
    except KeyboardInterrupt:
        pass
    finally:
        # Leave group and commit final offsets
        consumer.close()


def process_message(importer: EventsProcessor, msg: Message):
    """Process a message."""
    headers = msg.headers() or {}
    # Convert list of tuples to dict
    header_data = dict(headers)
    # Convert byte values to string values
    for key in header_data.keys():
        header_data[key] = header_data[key].decode("utf8")

    # Convert byte value to string value
    record_value = msg.value().decode("utf8")
    event_data = json.loads(record_value)

    logger.debug(msg.key())

    # Transform and load event into db
    importer.process_event(
        msg.key(),
        header_data,
        event_data,
    )



from argparse import ArgumentParser
from typing import Any, Dict, List, Tuple, Type

from django.apps import AppConfig, apps
from django.core.management import BaseCommand
from django.utils.functional import partition

from schematools.contrib.django.models import DynamicModel


class Command(BaseCommand):
    """Dump the dynamically generated ModelFactories."""

    help = "Dump the (dynamic) ModelMocker definitions that Django holds in-memory."  # noqa: A003

    model_meta_args: List[Tuple[str, Any]] = [
        # All possible options in Meta, with their defaults.
        # https://docs.djangoproject.com/en/3.2/ref/models/options/
        # The original model.Meta or Options.meta is not available after construction,
        # so will have to fill out the defaults to generate a reasonable representation.
        ("abstract", False),
        ("app_label", None),
        ("base_manager_name", None),
        ("constraints", []),
        ("db_table", None),
        ("default_manager_name", None),
        ("default_related_name", None),
        ("get_latest_by", None),
        ("index_together", ()),
        ("indexes", []),
        ("managed", True),
        ("order_with_respect_to", None),
        ("ordering", ["id"]),
        ("permissions", []),
        ("proxy", False),
        ("unique_together", ()),
        ("verbose_name", ""),  # is modelname
        ("verbose_name_plural", ""),  # is {modelname}s by default
    ]

    def add_arguments(self, parser: ArgumentParser) -> None:
        """Hook to add arguments."""
        parser.add_argument(
            "args",
            metavar="app_label",
            nargs="*",
            help="Names of Django apps to dump the ModelFactory for",
        )

    def handle(self, *args: str, **options: Any) -> None:
        """Main function of this command."""
        app_labels = sorted(args or apps.app_configs.keys())

        for app_label in app_labels:
            app: AppConfig = apps.get_app_config(app_label)
            self.write_header(app)
            models = app.get_models(include_auto_created=True)
            for model in sorted(models, key=lambda m: m._meta.model_name):
                # print(model)
                # <class 'dso_api.dynamic_api.aardgasvrijezones.models.buurt'>
                # <class 'dso_api.dynamic_api.aardgasvrijezones.models.buurtinitiatief'>
                self.write_model_mocker(model)

    def write_header(self, app: AppConfig) -> None:
        """Write app start header."""
        self.stdout.write(f"# ---- App: {app.verbose_name or app.label}\n\n\n")

    def write_model_mocker(self, model: Type[DynamicModel]) -> None:
        """Write the representation of a complete DjangoModelFactory (ModelMocker in our terminology) to the output."""
        bases = ", ".join(base_class.__name__ for base_class in model.__bases__)
        self.stdout.write(f"class {model.__name__}({bases}):\n")

        for field in model._meta.get_fields():
            print(repr(field))
        # <django.db.models.fields.BigIntegerField: id>
        # <django.contrib.gis.db.models.fields.GeometryField: geometry>
        # <django_postgres_unlimited_varchar.UnlimitedCharField: buurt_code>
        # <django_postgres_unlimited_varchar.UnlimitedCharField: buurt_naam>
        # <django_postgres_unlimited_varchar.UnlimitedCharField: toelichting>
        # <django_postgres_unlimited_varchar.UnlimitedCharField: aandeel_kookgas>
        # class buurtinitiatief(DynamicModel):
        # <django.db.models.fields.BigIntegerField: id>
        # <django.contrib.gis.db.models.fields.GeometryField: geometry>
        # <django_postgres_unlimited_varchar.UnlimitedCharField: buurtinitiatief_type>
        # <django_postgres_unlimited_varchar.UnlimitedCharField: buurt_naam>
        # <django.db.models.fields.FloatField: x_coordinaat>
        # <django.db.models.fields.FloatField: y_coordinaat>



from __future__ import annotations

from typing import Any, Collection, Dict, List, Optional, Tuple, Type, TypeVar
from urllib.parse import urlparse

from django.apps import apps
from django.conf import settings
from django.contrib.gis.db import models
from django.contrib.gis.db import models as gis_models
from django.contrib.postgres.fields import ArrayField
from django.db.models import CheckConstraint, Q
from django.db.models.base import ModelBase
from django_postgres_unlimited_varchar import UnlimitedCharField

from schematools import SRID_3D
from schematools.contrib.django import app_config, signals
from schematools.naming import to_snake_case, toCamelCase
from schematools.types import DatasetFieldSchema, DatasetTableSchema

from .mockers import DynamicModelMocker
from .models import (
    CompositeForeignKeyField,
    Dataset,
    DynamicModel,
    LooseRelationField,
    LooseRelationManyToManyField,
)

TypeAndSignature = Tuple[Type[models.Field], tuple, Dict[str, Any]]
M = TypeVar("M", bound=DynamicModel)
MODEL_CREATION_COUNTER = 1
MODEL_MOCKER_CREATION_COUNTER = 1

JSON_TYPE_TO_DJANGO = {
    "string": UnlimitedCharField,
    "integer": models.BigIntegerField,
    "integer/autoincrement": models.AutoField,
    "string/autoincrement": UnlimitedCharField,
    "datetime": models.DateTimeField,
    "number": models.FloatField,
    "boolean": models.BooleanField,
    "array": ArrayField,
    # Format variants of type string:
    "date": models.DateField,
    "time": models.TimeField,
    "date-time": models.DateTimeField,
    "uri": models.URLField,
    "email": models.EmailField,
    "blob-azure": UnlimitedCharField,
    # "object" handled elsewhere.
    "/definitions/id": models.IntegerField,
    "/definitions/schema": UnlimitedCharField,
    "https://geojson.org/schema/Geometry.json": gis_models.GeometryField,
    "https://geojson.org/schema/Point.json": gis_models.PointField,
    "https://geojson.org/schema/MultiPoint.json": gis_models.MultiPointField,
    "https://geojson.org/schema/Polygon.json": gis_models.PolygonField,
    "https://geojson.org/schema/MultiPolygon.json": gis_models.MultiPolygonField,
    "https://geojson.org/schema/LineString.json": gis_models.LineStringField,
    "https://geojson.org/schema/MultiLineString.json": gis_models.MultiLineStringField,
    "https://geojson.org/schema/GeometryCollection.json": gis_models.GeometryCollectionField,
}


def remove_dynamic_models() -> None:
    """Erase model caches for dynamically generated models.
    This completely removes the models from the Django app registry.

    If your own code also holds references to models, these need to be removed separately.
    The :func:`is_dangling_model` function allows to check whether a model originated
    from a previous factory invocation.
    """
    virtual_apps = [
        name
        for name, config in apps.app_configs.items()
        if isinstance(config, app_config.VirtualAppConfig)
    ]
    for app_label in virtual_apps:
        del apps.all_models[app_label]
        del apps.app_configs[app_label]

    # See if there are dynamic models registered in other apps, erase them too.
    for app_label, app_models in apps.all_models.items():
        dynamic_models = [
            name for name, model in app_models.items() if issubclass(model, DynamicModel)
        ]
        for model_name in dynamic_models:
            del app_models[model_name]
            del apps.app_configs[app_label].models[model_name]

    # Allow other apps to clear their caches too
    signals.dynamic_models_removed.send(sender=None)

    # This also clears FK caches of other models that may have foreign keys:
    apps.clear_cache()

    # Allow code to detect whether a model is still alive despite having cleared all caches.
    global MODEL_CREATION_COUNTER
    MODEL_CREATION_COUNTER += 1


def is_dangling_model(model: type[DynamicModel]) -> bool:
    """Tell whether the model should have been removed, as everything reloaded."""
    return model.CREATION_COUNTER < MODEL_CREATION_COUNTER


def schema_models_factory(
    dataset: Dataset,
    tables: Collection[str] | None = None,
    base_app_name: str | None = None,
    base_model: type[M] = DynamicModel,
) -> list[type[M]]:
    """Generate Django models from the data of the schema."""
    return [
        model_factory(
            dataset=dataset,
            table_schema=table,
            base_app_name=base_app_name,
            base_model=base_model,
        )
        for table in dataset.schema.get_tables(include_nested=True, include_through=True)
        if tables is None or table.id in tables
    ]


def schema_model_mockers_factory(
    dataset: Dataset,
    tables: Optional[Collection[str]] = None,
    base_app_name: Optional[str] = None,
) -> List[Type[DynamicModelMocker]]:
    """Generate Django model mockers from the data of the schema."""
    return [
        model_mocker_factory(dataset=dataset, table_schema=table, base_app_name=base_app_name)
        for table in dataset.schema.get_tables(include_nested=True, include_through=True)
        if tables is None or table.id in tables
    ]


def _get_model_name(table_schema: DatasetTableSchema) -> str:
    """Returns model name for this table. Including version number, if needed."""
    # Using table_schema.python_name gives UpperCamelCased names, the old format is kept here.
    # This also keeps model names more readable and recognizable/linkable with db table names.
    model_name = to_snake_case(table_schema.id)
    if table_schema.dataset.version is not None and not table_schema.dataset.is_default_version:
        return f"{model_name}_{table_schema.dataset.version}"
    else:
        return model_name


def model_factory(
    dataset: Dataset,
    table_schema: DatasetTableSchema,
    base_app_name: str | None = None,
    base_model: type[M] = DynamicModel,
    meta_options: dict[str, Any] | None = None,
) -> type[M]:
    """Generate a Django model class from a JSON Schema definition."""
    dataset_schema = dataset.schema
    app_label = dataset_schema.id
    base_app_name = base_app_name or "dso_api.dynamic_api"
    module_name = f"{base_app_name}.{app_label}.models"
    is_temporal = table_schema.is_temporal
    display_field = table_schema.display_field

    # Generate fields
    fields = {}
    constraints = []

    for field in table_schema.get_fields(include_subfields=True):
        # skip schema field for now
        # skip nested tables and fields that are only added for temporality
        if (
            field.type.endswith("definitions/schema")
            or field.is_nested_table
            or field.is_temporal_range
            or field.is_nested_object
        ):
            continue

        model_field = _model_field_factory(field)

        model_field.name = field.python_name  # Generate name, fix if needed.
        model_field.field_schema = field  # avoid extra lookups.
        fields[model_field.name] = model_field

        # Non-composite string identifiers may not contain forwardslashes, since this
        # breaks URL matching when they are used in URL paths.
        if field.is_primary and field.type == "string":
            # To make sure OneToOneField relations also support the '__contains' lookup,
            # that lookup type is registered with this field class in apps.py.
            constraints.append(
                CheckConstraint(
                    check=~Q(**{f"{model_field.name}__contains": "/"}),
                    name=f"{table_schema.db_name}_{field.db_name}_not_contains_slash",
                )
            )

    # Generate Meta part
    meta_cls = type(
        "Meta",
        (),
        {
            "managed": False,
            "db_table": table_schema.db_name,
            "app_label": app_label,
            "verbose_name": (table_schema.title or table_schema.id).capitalize(),
            "ordering": [idf.python_name for idf in table_schema.identifier_fields],
            "constraints": constraints,
            **(meta_options or {}),
        },
    )

    model_class = ModelBase(
        _get_model_name(table_schema),
        (base_model,),
        {
            **fields,
            "__doc__": table_schema.description or "",
            "_dataset": dataset,
            "_table_schema": table_schema,
            "_dataset_schema": dataset_schema,
            "_display_field": (display_field.python_name if display_field is not None else None),
            "_is_temporal": is_temporal,
            "CREATION_COUNTER": MODEL_CREATION_COUNTER,  # for debugging recreation
            "__module__": module_name,
            "Meta": meta_cls,
        },
    )

    app_config.register_model(app_label, model_class)
    return model_class


def _model_field_factory(field: DatasetFieldSchema) -> models.Field:
    """Construct the Django model field for a schema field."""
    if field.relation:
        return _fk_field_factory(field)
    elif field.nm_relation:
        return _nm_field_factory(field)
    else:
        return _basic_field_factory(field)


def _get_model_field_class(field: DatasetFieldSchema) -> type[models.Field]:
    type_ = field.type
    # reduce amsterdam schema refs to their fragment
    # only relevant for `/definitions/id` types atm.
    if type_.startswith(settings.SCHEMA_DEFS_URL):
        type_ = urlparse(type_).fragment

    try:
        return JSON_TYPE_TO_DJANGO[field.format or type_]
    except KeyError as e:
        raise RuntimeError(f"Field '{field.qualified_id}' has unsupported type: {type_}.") from e


def _get_basic_kwargs(field: DatasetFieldSchema) -> dict:
    """Common model field kwargs for all field types."""
    kwargs = {
        "primary_key": field.is_primary,
        "verbose_name": field.title,
        "help_text": field.description or "",  # also used by OpenAPI spec
        "db_column": field.db_name if field.db_name != field.python_name else None,
    }

    if not field.is_primary and field.nm_relation is None:
        # Primary can not be Null
        kwargs["null"] = not field.required

    return kwargs


def _basic_field_factory(field: DatasetFieldSchema) -> models.Field:
    """Construct a Django model field for a basic field type."""
    field_cls = _get_model_field_class(field)
    kwargs = _get_basic_kwargs(field)

    if issubclass(field_cls, ArrayField):
        # Array field
        item_type = field.get("entity", {}).get("type", "string")
        kwargs["base_field"] = JSON_TYPE_TO_DJANGO[item_type]()

    if issubclass(field_cls, gis_models.GeometryField):
        # Geometry field specials
        kwargs.update(
            {
                "srid": field.srid,
                "dim": (3 if field.srid in SRID_3D else 2),
                "geography": False,
                "db_index": True,
            }
        )

    return field_cls(**kwargs)


def _fk_field_factory(field: DatasetFieldSchema) -> models.ForeignKey:
    """Generate a Django ForeignKey field for a schema 1N relation field.

    This also takes composite-key relations into account,
    and "loose relations" where the field only references
    one part of the composite relation
    """
    to_field = _get_fk_to_field(field)
    kwargs = {
        "to": f"{field.related_table.dataset.id}.{_get_model_name(field.related_table)}",
        **_get_basic_kwargs(field),
        "on_delete": models.CASCADE if field.required else models.SET_NULL,
        "db_column": field.db_name,
        "db_constraint": False,  # don't enforce on database, not feasible for many datasets.
        "related_name": _get_fk_related_name(field),
    }

    if field.is_composite_key:
        # Make it easier to recognize the keys, e.g. in ``manage.py dump_models``.
        # For the most part, this is a tagging interface class.
        field_cls = CompositeForeignKeyField
        kwargs["to_fields"] = [field.python_name for field in field.related_fields]
    elif field.is_primary:
        # A primary key with a relation is typically a OneToOneField.
        # For a field named "id", using db_column="id" will work to retrieve the field there.
        # Internally, Django will still add an "id_id" field to the model to access the raw value.
        field_cls = models.OneToOneField
        kwargs["primary_key"] = True  # Note Django will use an INNER JOIN to retrieve the model.
        # Not using 'kwargs["parent_link"] = True', as model inheritance is not used here.
    elif field.is_loose_relation or to_field:
        # Points to the first part of a composite key (e.g. "identificatie").
        field_cls = LooseRelationField
        kwargs["to_field"] = to_field.python_name
    else:
        field_cls = models.ForeignKey

    return field_cls(**kwargs)


def _get_fk_related_name(field: DatasetFieldSchema) -> str:
    """Determine the name of the backwards relationship.

    If the linked table describes the other end of the relationship,
    this field will also be included in the model.
    """
    parent_table = field.table
    if parent_table.is_nested_table:
        # Won't ever show related name for internal tables
        return to_snake_case(parent_table["originalID"])
    elif parent_table.is_through_table:
        # This provides a reverse-link from each FK in the M2M table to the linked tables,
        # allowing to walk *inside* the M2M table instead of over it.
        # For debugging purposes these names are clarified depending on what direction
        # the key takes.
        m2m_field = parent_table.parent_table_field
        through_fields = parent_table["throughFields"]
        if field.name == through_fields[0]:
            # First model, can resemble the original field name,
            return f"rev_m2m_{to_snake_case(m2m_field.name)}"
        elif field.name == through_fields[1]:
            # Second model, can resemble the reverse name if it exists.
            m2m_reverse_name = m2m_field.reverse_relation
            if m2m_reverse_name is not None:
                # As the field exists on the second model,
                # let the reverse relation also reflect that.
                return f"rev_m2m_{to_snake_case(m2m_reverse_name.id)}"

        # By default, create something unique and recognizable.
        # The "m2m" would be snake_cased as m_2_m, so it's added afterwards.
        # The parent table ID already has the main field name included, but for tables
        # with a self-reference, the field is still added to guarantee uniqueness.
        return "rev_m2m_" + to_snake_case(f"{parent_table.id}_via_{field.id}")
    elif (additional_relation := field.reverse_relation) is not None:
        # The relation is described by the other table, return it
        return additional_relation.id
    else:
        # Hide it as relation.
        # Note that for M2M relations, Django will replace this with "_tablename_fieldname_+",
        # as Django still uses the backwards relation internally.
        return "+"


def _get_fk_to_field(field: DatasetFieldSchema) -> DatasetFieldSchema | None:
    """Determine the "to_field" for the foreign key.

    This returns a value when the relation doesn't point to the targets's primary key,
    hence the "to_field" parameter is needed.

    The current implementation only works for the right-side of N-M relations at the moment,
    other relation types are still created as a loose relation field.
    """
    if (
        # HACK: This complicated logic is needed because self.field.is_loose_relation
        # has very mixed-up logic that handles things which the callers should have handled.
        # This makes it impossible to determine whether a through-table has loose relations.
        # Solving that turns out to be really complex and bring up more issues. However,
        # reading the NM-field does work, so at least one side of the relation can be fixed.
        (nm_field := field.table.parent_table_field) is not None
        and nm_field.is_loose_relation
        and field.id == field.table["throughFields"][1]
    ):
        # A FK in the through table might actually be a partial composite key.
        # When this is the case, make sure the to_field points to the right field.
        target_fields = nm_field.related_fields

        if target_fields[0].id != "id" and not target_fields[0].is_primary:
            return target_fields[0]
    elif field.is_loose_relation:
        # Loose relation points to the first field of a composite foreign key
        return field.related_table.identifier_fields[0]

    return None


def _nm_field_factory(field: DatasetFieldSchema) -> models.ManyToManyField:
    """Generate a Django ManyToManyField for the NM-relation.

    NOTE: this still generates a regular M2M relation for "loose m2m relations".
    """
    # TODO: this doesn't really take loose relations into account.
    # In practice, the schematools still has 2 regular foreignkey pairs,
    # # but also includes subfields for the composite key field.
    field_cls = LooseRelationManyToManyField if field.is_loose_relation else models.ManyToManyField
    through_table = field.through_table
    through_fields = [f.python_name for f in through_table.through_fields]

    if (additional_relation := field.reverse_relation) is not None:
        # The relation is described by the other table, return it
        related_name = additional_relation.id
    else:
        # Default: give it a name, but hide it as relation.
        # This becomes the models.ManyToManyRel field on the target model.
        related_name = f"rev_{field.table.python_name}_{field.python_name}+"

    kwargs = {
        "to": f"{field.related_table.dataset.id}.{_get_model_name(field.related_table)}",
        **_get_basic_kwargs(field),
        "related_name": related_name,
        "through": f"{through_table.dataset.id}.{_get_model_name(through_table)}",
        "through_fields": through_fields,
    }

    return field_cls(**kwargs)


def _simplify_table_schema_relations(table_schema: DatasetTableSchema):
    """Remove relation attributes from relation definitions.

    This prevents the creation of Django FK and M2M fields,
    because we don't want to use those during mocking.
    """
    new_table_data = table_schema.json_data()
    new_table_data["schema"]["properties"] = {}
    # For FK add postfix and simplify the field, for M2M skip the field.
    for field in table_schema.get_fields(include_subfields=True):
        field_definition = field.json_data()
        # By using the db_name of the field, we are using
        # the fact the an `_id` postfix is added for relation fields.
        db_name = field.db_name
        # Some autogenerated schemas already have an `autoincrement` type.
        if field.is_primary and not field.data["type"].endswith("autoincrement"):
            field_definition["type"] += "/autoincrement"
        if field.nm_relation is not None:
            continue
        if field.relation is not None:
            # in rare cases, a relation field is part
            # of the identifier, so we need to remove it.
            if field.id in table_schema.identifier:
                new_table_data["schema"]["identifier"].remove(field.id)
            field_definition.pop("relation", None)
            field_definition.pop("shortname", None)
            # convert object-type fields to strings
            if field.type == "object":
                field_definition["type"] = "string"
                del field_definition["properties"]

            # Add autoincrementing behaviour to the identifier fields
            # by adding a `faker` attribute
            field_definition["faker"] = "nuller"
        # We need to convert the db_name back to the proper camelCasing
        # that is being used in amsterdam schemas.
        new_table_data["schema"]["properties"][toCamelCase(db_name)] = field_definition
    return DatasetTableSchema(new_table_data, parent_schema=table_schema._parent_schema)


def model_mocker_factory(
    dataset: Dataset, table_schema: DatasetTableSchema, base_app_name: Optional[str] = None
) -> Type[DynamicModelMocker]:
    """Generate a Django model mocker class from a JSON Schema definition."""
    # delayed import so Faker/shapely etc are not loaded for every application,
    # but only when mocker functionality is used.
    from .faker import get_field_factory

    dataset_schema = dataset.schema
    app_label = f"{dataset_schema.id}"
    base_app_name = base_app_name or "dso_api.dynamic_api"

    # Bootstrap for the model_factory is implemented in the DynamicRouter, so on startup
    # of the DSO-API. We have no desire to bootstrap the DynamicModelMocker for DSO runtime,
    # just in the test suite and in the relevant management commands. As the DynamicModelMocker
    # wraps around the DynamicModel, we must initiate it here to feed to the DjangoModelFactory
    # Meta class:
    # https://factoryboy.readthedocs.io/en/stable/orms.html#the-djangomodelfactory-subclass.
    # This table_schema is stripped from relation info, we want to use simple
    # scalars for relations.
    stripped_table_schema = _simplify_table_schema_relations(table_schema)
    model_factory(dataset, stripped_table_schema, base_app_name=base_app_name)

    # Generate fields
    fields = {}
    for field in stripped_table_schema.fields:
        type_ = field.type
        # skip schema field for now
        if type_.endswith("definitions/schema"):
            continue
        # skip nested tables and fields that are only added for temporality
        if field.is_nested_table or field.is_temporal_range:
            continue

        # Generate name, fix if needed.
        field_name = to_snake_case(field.name)
        fields[field_name] = get_field_factory(field)

    # Generate Meta part
    meta_cls = type(
        "Meta",
        (),
        {
            "model": f"{app_label}.{_get_model_name(stripped_table_schema)}",
            "database": "default",
        },
    )

    # The `Params` class in an inner class on the mocker,
    # it is needed during lazy evaluation, to provide schema information
    # during the mocking process.
    params_cls = type("Params", (), {"table_schema": table_schema})

    model_mocker_class = type(
        f"{_get_model_name(stripped_table_schema)}_factory",
        (DynamicModelMocker,),
        {
            **fields,
            "Meta": meta_cls,
            "Params": params_cls,
            "__module__": "schematools.contrib.django.factories",
        },
    )

    return model_mocker_class



# Generated by Django 3.1.13 on 2021-11-22 10:34

from django.db import migrations, models

from schematools.types import SemVer


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0013_profile_schema_data_as_textfield"),
    ]

    operations = [
        migrations.AddField(
            model_name="datasettable",
            name="version",
            field=models.TextField(default=SemVer("1.0.0")),
        ),
    ]



"""Very rudimentary implementation of fake geo data for now."""
import logging
import random

from django.contrib.gis import geos
from factory import Faker
from faker.providers import BaseProvider
from gisserver.geometries import CRS

from schematools import CRS_RD_NEW, SRID_3D, SRID_RD_NEW

LOCALE = "nl_NL"
DEFAULT_CRS = CRS_RD_NEW

# Amsterdam bounding box, in RD
WEST, SOUTH = 104000, 474000
EAST, NORTH = 136000, 501999

logger = logging.getLogger(__name__)


class UnsupportedGEOTypeException(Exception):
    """Exception to signal that the generic Geometry type is used.

    This type is not usable, because a faker does not know what type of geo data is needed.
    """


def random_point(has_z=False):
    point = (random.uniform(WEST, EAST), random.uniform(SOUTH, NORTH))  # noqa: S311  # nosec: B311
    return point if not has_z else point + (0.0,)


class Geometry:
    def __init__(self, has_z=False):
        class_name = self.__class__.__name__
        try:
            shape_cls = getattr(geos, class_name)
        except AttributeError:
            raise UnsupportedGEOTypeException(f"Class {class_name} is not allowed.")
        self.has_z = has_z
        self.shape = shape_cls(*self.get_coordinates(), srid=SRID_RD_NEW)

    def get_coordinates(self):
        return random_point(has_z=self.has_z)

    @property
    def ewkt(self):
        return self.shape.ewkt


class Point(Geometry):
    pass


class MultiPoint(Geometry):
    def get_coordinates(self):
        return [[random_point(has_z=self.has_z) for i in range(5)]]


class Polygon(Geometry):
    def get_coordinates(self):
        # Use a circle
        center = Point(has_z=self.has_z)
        return center.shape.buffer(1000)


class MultiPolygon(Geometry):
    def get_coordinates(self):
        return [[Polygon(has_z=self.has_z).shape for j in range(3)]]


class LineString(Geometry):
    def get_coordinates(self):
        return [[random_point(has_z=self.has_z) for i in range(5)]]


class MultiLineString(Geometry):
    def get_coordinates(self):
        return [[LineString(has_z=self.has_z).shape for j in range(3)]]


class GeometryCollection(Geometry):
    def get_coordinates(self):
        return [(Polygon(has_z=self.has_z).shape, LineString(has_z=self.has_z).shape)]


class GeoProvider(BaseProvider):  # noqa: D101
    """Geoprovider that selects a concrete implementation class."""

    def geo(self, geo_type: str = "Geometry", crs: CRS = DEFAULT_CRS) -> str:  # noqa: D102
        try:
            shape = globals()[geo_type](has_z=crs.srid in SRID_3D)
            if crs != DEFAULT_CRS:
                crs.apply_to(shape.shape)
            ewkt = shape.ewkt
        except (UnsupportedGEOTypeException, KeyError, TypeError) as e:
            message = str(e)
            logger.warning("Skipping geo data creation, reason: %s", message)
            return None
        else:
            return ewkt


Faker.add_provider(GeoProvider, locale=LOCALE)



# Generated by Django 3.0.3 on 2020-02-13 06:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0002_add_datasettable"),
    ]

    operations = [
        migrations.AddField(
            model_name="datasettable",
            name="geometry_field_type",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
    ]
